"""
Sub-controllers for navigate.
"""

from .stage_controller import StageController  # noqa
from .acquire_bar_controller import AcquireBarController  # noqa
from .channels_tab_controller import ChannelsTabController  # noqa
from .camera_view_controller import CameraViewController  # noqa
from .camera_setting_controller import CameraSettingController  # noqa
from .waveform_tab_controller import WaveformTabController  # noqa
from .waveform_popup_controller import WaveformPopupController  # noqa
from .autofocus_popup_controller import AutofocusPopupController  # noqa
from .features_popup_controller import FeaturePopupController  # noqa
from .feature_advanced_setting_controller import (
    FeatureAdvancedSettingController,  # noqa
)
from .keystroke_controller import KeystrokeController  # noqa
from .multi_position_controller import MultiPositionController  # noqa
from .ilastik_popup_controller import IlastikPopupController  # noqa
from .camera_map_setting_popup_controller import CameraMapSettingPopupController  # noqa
from .microscope_popup_controller import MicroscopePopupController  # noqa
from .adaptiveoptics_popup_controller import AdaptiveOpticsPopupController  # noqa
# from .uninstall_plugin_controller import UninstallPluginController  # noqa
from .plugins_controller import PluginsController, UninstallPluginController  # noqa
from .menu_controller import MenuController  # noqa